# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: topgen/payload.proto
# Protobuf Python Version: 5.28.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    0,
    '',
    'topgen/payload.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14topgen/payload.proto\"\xdf\x04\n\x07Payload\x12.\n\x0eotp_parameters\x18\x01 \x03(\x0b\x32\x16.Payload.OtpParameters\x12\x0f\n\x07version\x18\x02 \x01(\x05\x12\x12\n\nbatch_size\x18\x03 \x01(\x05\x12\x13\n\x0b\x62\x61tch_index\x18\x04 \x01(\x05\x12\x10\n\x08\x62\x61tch_id\x18\x05 \x01(\x05\x1a\xba\x01\n\rOtpParameters\x12\x0e\n\x06secret\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0e\n\x06issuer\x18\x03 \x01(\t\x12%\n\talgorithm\x18\x04 \x01(\x0e\x32\x12.Payload.Algorithm\x12#\n\x06\x64igits\x18\x05 \x01(\x0e\x32\x13.Payload.DigitCount\x12\x1e\n\x04type\x18\x06 \x01(\x0e\x32\x10.Payload.OtpType\x12\x0f\n\x07\x63ounter\x18\x07 \x01(\x04\"y\n\tAlgorithm\x12\x19\n\x15\x41LGORITHM_UNSPECIFIED\x10\x00\x12\x12\n\x0e\x41LGORITHM_SHA1\x10\x01\x12\x14\n\x10\x41LGORITHM_SHA256\x10\x02\x12\x14\n\x10\x41LGORITHM_SHA512\x10\x03\x12\x11\n\rALGORITHM_MD5\x10\x04\"U\n\nDigitCount\x12\x1b\n\x17\x44IGIT_COUNT_UNSPECIFIED\x10\x00\x12\x13\n\x0f\x44IGIT_COUNT_SIX\x10\x01\x12\x15\n\x11\x44IGIT_COUNT_EIGHT\x10\x02\"I\n\x07OtpType\x12\x18\n\x14OTP_TYPE_UNSPECIFIED\x10\x00\x12\x11\n\rOTP_TYPE_HOTP\x10\x01\x12\x11\n\rOTP_TYPE_TOTP\x10\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'topgen.payload_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PAYLOAD']._serialized_start=25
  _globals['_PAYLOAD']._serialized_end=632
  _globals['_PAYLOAD_OTPPARAMETERS']._serialized_start=161
  _globals['_PAYLOAD_OTPPARAMETERS']._serialized_end=347
  _globals['_PAYLOAD_ALGORITHM']._serialized_start=349
  _globals['_PAYLOAD_ALGORITHM']._serialized_end=470
  _globals['_PAYLOAD_DIGITCOUNT']._serialized_start=472
  _globals['_PAYLOAD_DIGITCOUNT']._serialized_end=557
  _globals['_PAYLOAD_OTPTYPE']._serialized_start=559
  _globals['_PAYLOAD_OTPTYPE']._serialized_end=632
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Payload(_message.Message):
    __slots__ = ("otp_parameters", "version", "batch_size", "batch_index", "batch_id")
    class Algorithm(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        ALGORITHM_UNSPECIFIED: _ClassVar[Payload.Algorithm]
        ALGORITHM_SHA1: _ClassVar[Payload.Algorithm]
        ALGORITHM_SHA256: _ClassVar[Payload.Algorithm]
        ALGORITHM_SHA512: _ClassVar[Payload.Algorithm]
        ALGORITHM_MD5: _ClassVar[Payload.Algorithm]
    ALGORITHM_UNSPECIFIED: Payload.Algorithm
    ALGORITHM_SHA1: Payload.Algorithm
    ALGORITHM_SHA256: Payload.Algorithm
    ALGORITHM_SHA512: Payload.Algorithm
    ALGORITHM_MD5: Payload.Algorithm
    class DigitCount(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        DIGIT_COUNT_UNSPECIFIED: _ClassVar[Payload.DigitCount]
        DIGIT_COUNT_SIX: _ClassVar[Payload.DigitCount]
        DIGIT_COUNT_EIGHT: _ClassVar[Payload.DigitCount]
    DIGIT_COUNT_UNSPECIFIED: Payload.DigitCount
    DIGIT_COUNT_SIX: Payload.DigitCount
    DIGIT_COUNT_EIGHT: Payload.DigitCount
    class OtpType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        OTP_TYPE_UNSPECIFIED: _ClassVar[Payload.OtpType]
        OTP_TYPE_HOTP: _ClassVar[Payload.OtpType]
        OTP_TYPE_TOTP: _ClassVar[Payload.OtpType]
    OTP_TYPE_UNSPECIFIED: Payload.OtpType
    OTP_TYPE_HOTP: Payload.OtpType
    OTP_TYPE_TOTP: Payload.OtpType
    class OtpParameters(_message.Message):
        __slots__ = ("secret", "name", "issuer", "algorithm", "digits", "type", "counter")
        SECRET_FIELD_NUMBER: _ClassVar[int]
        NAME_FIELD_NUMBER: _ClassVar[int]
        ISSUER_FIELD_NUMBER: _ClassVar[int]
        ALGORITHM_FIELD_NUMBER: _ClassVar[int]
        DIGITS_FIELD_NUMBER: _ClassVar[int]
        TYPE_FIELD_NUMBER: _ClassVar[int]
        COUNTER_FIELD_NUMBER: _ClassVar[int]
        secret: bytes
        name: str
        issuer: str
        algorithm: Payload.Algorithm
        digits: Payload.DigitCount
        type: Payload.OtpType
        counter: int
        def __init__(self, secret: _Optional[bytes] = ..., name: _Optional[str] = ..., issuer: _Optional[str] = ..., algorithm: _Optional[_Union[Payload.Algorithm, str]] = ..., digits: _Optional[_Union[Payload.DigitCount, str]] = ..., type: _Optional[_Union[Payload.OtpType, str]] = ..., counter: _Optional[int] = ...) -> None: ...
    OTP_PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    BATCH_SIZE_FIELD_NUMBER: _ClassVar[int]
    BATCH_INDEX_FIELD_NUMBER: _ClassVar[int]
    BATCH_ID_FIELD_NUMBER: _ClassVar[int]
    otp_parameters: _containers.RepeatedCompositeFieldContainer[Payload.OtpParameters]
    version: int
    batch_size: int
    batch_index: int
    batch_id: int
    def __init__(self, otp_parameters: _Optional[_Iterable[_Union[Payload.OtpParameters, _Mapping]]] = ..., version: _Optional[int] = ..., batch_size: _Optional[int] = ..., batch_index: _Optional[int] = ..., batch_id: _Optional[int] = ...) -> None: ...



import argparse
import base64
import hmac
import os
import time
import urllib.parse

import tqdm

import topgen.payload_pb2 as payload_pb2

INTERVAL = 30


# https://stackoverflow.com/questions/8529265/google-authenticator-implementation-in-python
# https://stackoverflow.com/a/69343235/7075029
# TOTP: Time-Based One-Time Password Algorithm: https://www.rfc-editor.org/rfc/rfc6238
# HOTP: An HMAC-Based One-Time Password Algorithm: https://www.rfc-editor.org/rfc/rfc4226
def totp(key: bytes, counter: int):
    msg = counter.to_bytes(8, "big")
    digest = hmac.new(key, msg, "sha1").digest()
    offset = digest[19] & 0xF
    code = digest[offset : offset + 4]
    code = int.from_bytes(code, "big") & 0x7FFFFFFF
    code = code % 1000000
    return "{:06d}".format(code)


def list_items(payload):
    for i, p in enumerate(payload.otp_parameters):
        print(f"#{i}")
        print_item(p)


def print_item(p):
    print("Secret:", base64.b32encode(p.secret).decode())
    print("Name: ", p.name)
    print("Issuer: ", p.issuer)
    print("Algorithm: ", p.algorithm)
    print("Digits: ", p.digits)
    print("Type: ", p.type)
    print("Counter: ", p.counter)
    print()


def decode_url(url):
    url = urllib.parse.urlparse(url)
    data = urllib.parse.parse_qs(url.query)["data"][0]
    bin = base64.b64decode(data)
    payload = payload_pb2.Payload()
    payload.ParseFromString(bin)
    return payload


def save(filename, payload):
    with open(filename, "wb") as f:
        f.write(payload.SerializeToString())


def read(filename, payload):
    with open(filename, "rb") as f:
        payload.ParseFromString(f.read())


def get_name(p):
    return f"{p.issuer} ({p.name})"


def main(args):
    try:
        payload = payload_pb2.Payload()
        if os.path.exists(args.db):
            read(args.db, payload)
        if args.url:
            payload.MergeFrom(decode_url(args.url))
            save(args.db, payload)

        if args.add:
            secret = input("Secret > ").strip().replace(" ", "")
            try:
                secret = base64.b32decode(secret, casefold=True)
            except Exception:
                print(f'Given secret "{secret}" is not valid. Secret must be base32 decodeable.')
                return
            name = input("Name > ")
            issuer = input("Issuer > ")
            p = payload.otp_parameters.add()
            p.secret = secret
            p.name = name
            p.issuer = issuer
            p.algorithm = payload_pb2.Payload.Algorithm.ALGORITHM_SHA1
            p.digits = payload_pb2.Payload.DigitCount.DIGIT_COUNT_SIX
            p.type = payload_pb2.Payload.OtpType.OTP_TYPE_TOTP
            save(args.db, payload)

        payload.otp_parameters.sort(key=lambda p: get_name(p))
        num_items = len(payload.otp_parameters)
        if num_items > 0:
            if args.remove:
                list_items(payload)
                index = int(input(f"Select index to delete [0-{num_items - 1}]> "))
                if index < 0 or index >= num_items:
                    print(f"Index have to be between 0-{num_items - 1}")
                    return
                print_item(payload.otp_parameters[index])
                answer = input("Are you sure to delete above item? [y,N] > ")
                if answer == "y":
                    del payload.otp_parameters[index]
                    save(args.db, payload)
                    print("Item has been deleted")
                else:
                    print("Item has not been deleted")
                return

            if args.list:
                list_items(payload)
                return

            while True:
                counter = int(time.time() // INTERVAL)
                print()
                for p in payload.otp_parameters:
                    print(f"{get_name(p)}: {totp(p.secret, counter)}")
                print()
                with tqdm.tqdm(total=INTERVAL, bar_format="{bar}") as progress:
                    prev_time = 0
                    while int(time.time() % INTERVAL) > 0:
                        curr_time = time.time() % INTERVAL
                        progress.update(curr_time - prev_time)
                        prev_time = curr_time
                        time.sleep(0.5)
                time.sleep(1)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    parser = argparse.ArgumentParser("TOPGen - Time-based one-time password generator")
    parser.add_argument("--db", required=False, default="payload.db", help="file path to store keys")
    parser.add_argument(
        "--url",
        required=False,
        help='URL in the form of "otpauth-migration://offline?data=..." to import keys from',
    )
    parser.add_argument("--list", action="store_true", help="list all the keys with their details")
    parser.add_argument("--add", action="store_true", help="add a new key")
    parser.add_argument("--remove", action="store_true", help="remove a key")
    args = parser.parse_args()
    main(args)
